/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in 
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2007.  All Rights Reserved.
 *
 * Protege was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu.
 *
 */


package edu.stanford.smi.protegex.owl.swrl.bridge.builtins.swrlxml;

import edu.stanford.smi.protegex.owl.swrl.bridge.ArgumentFactory;
import edu.stanford.smi.protegex.owl.swrl.bridge.OWLFactory;
import edu.stanford.smi.protegex.owl.swrl.bridge.BuiltInArgument;
import edu.stanford.smi.protegex.owl.swrl.bridge.MultiArgument;
import edu.stanford.smi.protegex.owl.swrl.bridge.OWLIndividual;
import edu.stanford.smi.protegex.owl.swrl.bridge.exceptions.InvalidBuiltInArgumentException;
import edu.stanford.smi.protegex.owl.swrl.bridge.exceptions.BuiltInException;
import edu.stanford.smi.protegex.owl.swrl.bridge.exceptions.SWRLRuleEngineBridgeException;
import edu.stanford.smi.protegex.owl.swrl.bridge.builtins.AbstractSWRLBuiltInLibrary;
import edu.stanford.smi.protegex.owl.swrl.bridge.builtins.SWRLBuiltInUtil;
import edu.stanford.smi.protegex.owl.swrl.util.SWRLOWLUtil;
import edu.stanford.smi.protegex.owl.swrl.exceptions.SWRLOWLUtilException;

import edu.stanford.smi.protegex.owl.model.OWLModel;

import org.jdom.*;
import org.jdom.xpath.*;

import java.util.*;

/**
 */
public class SWRLBuiltInLibraryImpl extends AbstractSWRLBuiltInLibrary 
{
  private static String SWRLXMLLibraryName = "SWRLXMLBuiltIns";

  private ArgumentFactory argumentFactory;

  public SWRLBuiltInLibraryImpl() 
  { 
    super(SWRLXMLLibraryName); 

    reset();

    argumentFactory = ArgumentFactory.getFactory();
  } // SWRLBuiltInLibraryImpl

  private XMLProcessor xmlProcessor;
  private XMLMapper xmlMapper;

  private Map<String, OWLIndividual> documentMappings; // File name to OWL document individuals
  private Map<String, Document> documents; // Individual name to Document

  private Map<String, Set<OWLIndividual>> elementMappings; // XML path to element individuals
  private Map<String, Element> elements; // Individual name to Element

  public void reset() 
  {
    xmlProcessor = new XMLProcessor();
    xmlMapper = new XMLMapper();

    documentMappings = new HashMap<String, OWLIndividual>();
    documents = new HashMap<String, Document>();

    elementMappings = new HashMap<String, Set<OWLIndividual>>();
    elements = new HashMap<String, Element>();
  } // reset

  public boolean makeXMLDocument(List<BuiltInArgument> arguments) throws BuiltInException 
  { 
    boolean result = false;

    SWRLBuiltInUtil.checkNumberOfArgumentsEqualTo(2, arguments.size());
    SWRLBuiltInUtil.checkForUnboundNonFirstArguments(arguments);

    if (SWRLBuiltInUtil.isUnboundArgument(0, arguments)) {
      String inputXMLStreamName = SWRLBuiltInUtil.getArgumentAsAString(1, arguments);
      OWLIndividual xmlDocument = null;
      Document doc;

      if (!documentMappings.containsKey(inputXMLStreamName)) {

        try {
          doc = xmlProcessor.processXMLStream(inputXMLStreamName);
          xmlDocument = xmlMapper.document2XMLDocumentMapping(doc, getInvokingBridge());
          documentMappings.put(inputXMLStreamName, xmlDocument);
          documents.put(xmlDocument.getIndividualName(), doc);
        } catch (XMLProcessorException e) {
          throw new BuiltInException("error processing XML stream '" + inputXMLStreamName + "': " + e.getMessage());
        } catch (XMLMapperException e) {
          throw new BuiltInException("error mapping XML stream '" + inputXMLStreamName + "': " + e.getMessage());
        } // try
      } else xmlDocument = documentMappings.get(inputXMLStreamName);

      arguments.set(0, xmlDocument); // Bind the result to the first parameter
      result = true;
    } // if
    return result;
  } // makeXMLDocument

  public boolean element(List<BuiltInArgument> arguments) throws BuiltInException 
  {
    boolean result = false;

    SWRLBuiltInUtil.checkNumberOfArgumentsEqualTo(3, arguments.size());
    SWRLBuiltInUtil.checkForUnboundNonFirstArguments(arguments);

    if (SWRLBuiltInUtil.isUnboundArgument(0, arguments)) {
      Document doc = getArgumentAsADocument(1, arguments);
      String path = SWRLBuiltInUtil.getArgumentAsAString(2, arguments);
      Set<OWLIndividual> owlElements;

      if (!elementMappings.containsKey(path)) {

        try {
          XPath xPath = XPath.newInstance(path);
          Iterator elementIterator = xPath.selectNodes(doc).iterator();
          owlElements = new HashSet<OWLIndividual>();
          
          while (elementIterator.hasNext()) {
            Object o = elementIterator.next();
            OWLIndividual xmlElement = getInvokingBridge().createOWLIndividual(OWLFactory.createOWLClass(XMLMapper.XMLElementMappingOWLClassName));
            Element element;
            
            if (!(o instanceof Element)) 
              throw new BuiltInException("path '" + path + "' must only refer to XML elements, found '" + o.getClass().getName() + "'");

            element = (Element)o;
            elements.put(xmlElement.getIndividualName(), element);

            owlElements.add(xmlElement);
          } // while
          if (!owlElements.isEmpty()) elementMappings.put(path, owlElements);
        } catch (JDOMException e) {
          throw new BuiltInException("JDOM error processing XML path '" + path + "': " + e.getMessage());
        } catch (SWRLRuleEngineBridgeException e) {
          throw new BuiltInException("error creating individual when processing XML path '" + path + "': " + e.getMessage());
        } // try
      } else owlElements = elementMappings.get(path);

      if (owlElements.isEmpty()) result = false;
      else if (owlElements.size() == 1) {
        arguments.set(0, (OWLIndividual)owlElements.toArray()[0]); // Bind the single individual to the first parameter
        result = true;
      } else {
        MultiArgument multiArgument = argumentFactory.createMultiArgument(SWRLBuiltInUtil.getVariableName(0, arguments), 
                                                                          SWRLBuiltInUtil.getPrefixedVariableName(0, arguments));
        for (OWLIndividual xmlElement : owlElements) 
          multiArgument.addArgument(argumentFactory.createIndividualArgument(xmlElement.getIndividualName()));
        arguments.set(0, multiArgument);
        result = !multiArgument.hasNoArguments();
      } // if
    } // if
    return result;
  } // element

  public boolean subElement(List<BuiltInArgument> arguments) throws BuiltInException 
  {
    boolean result = false;

    SWRLBuiltInUtil.checkNumberOfArgumentsEqualTo(3, arguments.size());
    SWRLBuiltInUtil.checkForUnboundNonFirstArguments(arguments);

    if (SWRLBuiltInUtil.isUnboundArgument(0, arguments)) {
      Element parent = getArgumentAsAnElement(1, arguments);
      String path = SWRLBuiltInUtil.getArgumentAsAString(2, arguments);
      Set<OWLIndividual> owlElements;

      try {
        XPath xPath = XPath.newInstance(path);
        Iterator elementIterator = xPath.selectNodes(parent).iterator();
        owlElements = new HashSet<OWLIndividual>();
        
        while (elementIterator.hasNext()) {
          Object o = elementIterator.next();
          OWLIndividual xmlElement = getInvokingBridge().createOWLIndividual(OWLFactory.createOWLClass(XMLMapper.XMLElementMappingOWLClassName));
          Element element;
          
          if (!(o instanceof Element)) 
            throw new BuiltInException("path '" + path + "' must only refer to XML elements, found '" + o.getClass().getName() + "'");
          
          element = (Element)o;
          elements.put(xmlElement.getIndividualName(), element);
          
          owlElements.add(xmlElement);
        } // while
      } catch (JDOMException e) {
        throw new BuiltInException("JDOM error processing XML path '" + path + "': " + e.getMessage());
      } catch (SWRLRuleEngineBridgeException e) {
        throw new BuiltInException("error creating individual when processing XML path '" + path + "': " + e.getMessage());
      } // try
      
      if (owlElements.isEmpty()) result = false;
      else if (owlElements.size() == 1) {
        arguments.set(0, (OWLIndividual)owlElements.toArray()[0]); // Bind the single individual to the first parameter
        result = true;
      } else {
        MultiArgument multiArgument = argumentFactory.createMultiArgument(SWRLBuiltInUtil.getVariableName(0, arguments), 
                                                                          SWRLBuiltInUtil.getPrefixedVariableName(0, arguments));
        for (OWLIndividual xmlElement : owlElements) 
          multiArgument.addArgument(argumentFactory.createIndividualArgument(xmlElement.getIndividualName()));
        arguments.set(0, multiArgument);
        result = !multiArgument.hasNoArguments();
      } // if
    } // if
    return result;
  } // element


  public boolean attributeValue(List<BuiltInArgument> arguments) throws BuiltInException 
  {
    boolean result = false;

    SWRLBuiltInUtil.checkNumberOfArgumentsEqualTo(3, arguments.size());
    SWRLBuiltInUtil.checkForUnboundNonFirstArguments(arguments);

    if (SWRLBuiltInUtil.isUnboundArgument(0, arguments)) {
      Element element = getArgumentAsAnElement(1, arguments);
      String attributeName = SWRLBuiltInUtil.getArgumentAsAString(2, arguments);
      String attributeValue = element.getAttributeValue(attributeName);

      if (attributeValue != null) {
        arguments.set(0, argumentFactory.createDatatypeValueArgument(attributeValue));
        result = true;
      } // if
    } // if
    return result;
  } // attributeValue

  private Document getArgumentAsADocument(int argumentNumber, List<BuiltInArgument> arguments) throws BuiltInException
  {
    Document document = null;

    if (SWRLBuiltInUtil.isArgumentAnIndividual(argumentNumber, arguments)) {
      String individualName = SWRLBuiltInUtil.getArgumentAsAnIndividualName(argumentNumber, arguments);
      
      if (isIndividualOfClass(individualName, XMLMapper.XMLDocumentMappingOWLClassName)) {
      } else throw new InvalidBuiltInArgumentException(argumentNumber, "individual '" + individualName + "' is not a " +
                                                       XMLMapper.XMLDocumentMappingOWLClassName);

      if (documents.containsKey(individualName)) document = documents.get(individualName);
      else throw new InvalidBuiltInArgumentException(argumentNumber, "" + XMLMapper.XMLDocumentMappingOWLClassName + " individual '" + 
                                                     individualName + "' does not refer to a valid document");
    } else throw new InvalidBuiltInArgumentException(argumentNumber, "expecting a " +  XMLMapper.XMLDocumentMappingOWLClassName + " individual" +
                                                     ", got '" + arguments.get(argumentNumber) + "'");
    return document;
  } // getArgumentAsADocument

  private Element getArgumentAsAnElement(int argumentNumber, List<BuiltInArgument> arguments) throws BuiltInException
  {
    Element element = null;

    if (SWRLBuiltInUtil.isArgumentAnIndividual(argumentNumber, arguments)) {
      String individualName = SWRLBuiltInUtil.getArgumentAsAnIndividualName(argumentNumber, arguments);
      
      if (isIndividualOfClass(individualName, XMLMapper.XMLElementMappingOWLClassName)) {
      } else throw new InvalidBuiltInArgumentException(argumentNumber, "individual '" + individualName + "' is not a " +
                                                       XMLMapper.XMLElementMappingOWLClassName);

      if (elements.containsKey(individualName)) element = elements.get(individualName);
      else throw new InvalidBuiltInArgumentException(argumentNumber, "" + XMLMapper.XMLElementMappingOWLClassName + " individual '" + 
                                                     individualName + "' does not refer to a valid element");
    } else throw new InvalidBuiltInArgumentException(argumentNumber, "expecting a " +  XMLMapper.XMLElementMappingOWLClassName + " individual" +
                                                     ", got '" + arguments.get(argumentNumber) + "'");
    return element;
  } // getArgumentAsAnElement

  private boolean isIndividualOfClass(String individualName, String className) throws BuiltInException
  {
    boolean result = false;

    try {
      result = SWRLOWLUtil.isIndividualOfClass(getInvokingBridge().getOWLModel(), individualName, className);
    } catch (SWRLOWLUtilException e) {
      throw new BuiltInException("internal XML processing error: " + e.getMessage());
    } // try
    return result;
  } // isIndividualOfClass

} // SWRLBuiltInLibraryImpl
